import numpy as np
from sinnn.utils import sigmoid, softmax, one_hot


class CrossEntropy:
    """Calculates the rate of change of loss with respect to logits.

    Attributes
    ----------
    isbinary : bool
        Whether CrossEntropy should calculate for binary or categorical CrossEntropy.

    """

    def __init__(self, isbinary=None):
        """
        Initialises Crossentropy.

        Parameters
        ----------
        isbinary : bool
            Whether the loss function should calculate gradient of Binary CrossEntropy or Categorical CrossEntropy.
            Initialised to None by default.
        """

        self.isbinary = isbinary
        pass

    def loss_gradient(self, z, y):
        """
        Calculates dL/dz where L represents Loss and z represents logits

        Parameters
        ----------
        z : ndarray
            Array containing logits of the current batch.
        y : ndArray
            Array containing labels of the current batch

        Returns
        -------
        ndarray
            Array of same shape as z and y containing dL/dz.

        Notes
        -----

        Binary CrossEntropy:
        since L = Σ 1/n -(ylog(σ(z)) + (1-y)log(1 - σ(z)))
        where:
            L = loss
            n = length of 1d array `z`
            y = labels
            log = natural logarithm
            σ = sigmoid function
            z = logits
        dL/dz = Σ 1/n -(y(σ(z))'/(σ(z)) + (1-y)(1 - σ(z))'/(1 - σ(z)))
              = Σ 1/n -(y(σ(z))(1 - σ(z))/(σ(z)) + (1-y)(-(σ(z))(1 - σ(z)))/(1 - σ(z))) since (σ(z))' = (σ(z))(1 - σ(z))
              = Σ 1/n -(y(1 - σ(z)) - (1-y)(σ(z)))
              = Σ 1/n -(y - y(σ(z)) - σ(z) + y(σ(z)))
              = Σ 1/n -(y  - σ(z))
              = Σ 1/n (σ(z) - y)
        the entire array is returned instead of the summation, as it will be dotted later.

        Categorical CrossEntropy:
        please see:
            https://deepnotes.io/softmax-crossentropy#derivative-of-cross-entropy-loss-with-softmax [1]

        the entire array is returned instead of the summation, as it will be dotted later.

        [1] P. Dahal, "Classification and Loss Evaluation - Softmax and Cross Entropy Loss", DeepNotes, 2020. [Online].
        Available: https://deepnotes.io/softmax-crossentropy#derivative-of-cross-entropy-loss-with-softmax.
        [Accessed: 16- Nov- 2020]

        """

        if self.isbinary and np.any(y > 1):
            raise Exception("Can not use 1 neuron for output layer of categorical classification task")
        return (sigmoid(z) - y)/len(z) if self.isbinary else (softmax(z) - one_hot(y, z.shape[1]))/len(z)


class MSE:
    """Calculates the rate of change of loss with respect to logits."""

    def __init__(self, isbinary=None):
        """
        Initialises Crossentropy.

        Parameters
        ----------
        isbinary : bool
            Initialised to None by default.

        """

        self.isbinary = isbinary
        pass

    def loss_gradient(self, z, y):
        """
        Calculates dL/dz where L represents Loss and z represents logits

        Parameters
        ----------
        z : ndarray
            1d Array containing logits of the current batch.
        y : ndArray
            1d Array containing labels of the current batch

        Returns
        -------
        ndarray
            1d Array of same shape as z and y containing dL/dz.

        Notes
        -----
        since L = 1/n Σ(y - z)^2
        where:
            L = loss
            n = length of 1d array `z`
            y = labels
            z = logits
        dL/dz = Σ 1/n 2*(y-z)*(-1)
              = Σ 2/n * (z-y)
        the entire array is returned instead of the summation, as it will be dotted later.

        """

        return 2/len(z) * (z-y)




import numpy as np


class Zeros:
    """Matrix of zeros."""
    def __init__(self):
        pass

    def initialise(self, rows, columns):
        """Returns a matrix of zeros.

        Parameters
        ----------
        rows : (int)
            The number of rows in the resulting matrix.
        columns : (int)
            The number of columns of the resulting matrix.

        Returns
        -------
        ndarray
            Array of zeros with shape: (`rows`, `columns`)

        """

        return np.zeros((rows, columns))


class SmallRandom:
    """ Matrix of random floats in the range 0 - lim"""

    def __init__(self):
        pass

    def initialise(self, rows, columns, lim=0.01):
        """Returns a matrix of random numbers in the range 0 - lim.

        Parameters
        ----------
        rows : (int)
            The number of rows in the resulting matrix.
        columns : (int)
            The number of columns of the resulting matrix.
        lim : (float)
            The upper limit of the random numbers.

        Returns
        -------
        ndarray
            Array of random numbers in range 0 - `lim` with shape: (`rows`, `columns`)

        """

        return np.random.randn(rows, columns) * lim



import numpy as np

class SGD:
    """Implements stochastic gradient descent."""

    def __init__(self, learning_rate=0.01):
        """
        Initialises SGD parameters.

        Parameters
        ----------
        learning_rate : float
            learning rate of the optimiser
        """
        self.learning_rate = learning_rate

    def reset(self):
        pass

    def optimise(self, weights, grad_weights):
        """
        Moves the weights in the direction opposite to weights gradient to attempt to reduce the Loss.

        Stochastic gradient descent is used to make small adjustments to the weights in an attempt to minimize the Loss.
        The size of the adjustments are determined by the learning rate.

        Parameters
        ----------
        weights : ndarray
            The existing weights which the optimiser needs to change.
        grad_weights : ndarray
            The weights gradient that determine which direction do the `weights` move

        Returns
        -------
        ndarray
            The new weights

        Notes
        -----
        w1 = w0 - n(dL/dw)
        where:
            w1 = the new weights
            w0 = The existing weights
            n = learning rate
            dL/dw = The weights gradient

        """
        return weights - self.learning_rate * grad_weights


class Momentum:
    """ Implements stochastic gradient descent with Momentum. """

    def __init__(self, learning_rate=0.01, eta=0.9):
        """
        Initialises Momentum parameters.

        Parameters
        ----------
        learning_rate : float
            learning rate of the optimiser
        eta : float
            momentum multiplier

        """

        self.learning_rate = learning_rate
        self.eta = eta
        # Momentum vector initialised to None
        self.momentum = None

    def reset(self):
        """Resets momentum to 0"""
        self.momentum = 0

    def optimise(self, weights, grad_weights):
        """
        Moves the weights in the direction opposite to weights gradient and momentum to attempt to reduce the Loss.

        Momentum is an extension of Stochastic gradient descent. A momentum vector is kept in memory
        which also determines the direction of adjustments to weights, other than weight gradients.
        The momentum vector is determined by weight gradients and previous momentum.

        Parameters
        ----------
        weights : ndarray
            The existing weights which the optimiser needs to change.
        grad_weights : ndarray
            The weights gradient that determine which direction do the `weights` move

        Returns
        -------
        ndarray
            The new weights

        Notes
        -----
        h1 = a(h0) + n(dL/dw)
        w1 = w0 - h1

        where:
            w1 = the new weights
            w0 = The existing weights
            n = learning rate
            dL/dw = The weights gradient
            h1 = the new momentum
            h0 = the previous momentum

        """

        self.momentum = self.eta * self.momentum + self.learning_rate * grad_weights
        return weights - self.momentum


class AdaGrad:
    """
    Initialises AdaGrad parameters.

    Parameters
    ----------
    learning_rate : float
        learning rate of the optimiser

    """

    def __init__(self, learning_rate=0.01):
        self.learning_rate = learning_rate
        # g vector initialised to None
        self.g = None

    def reset(self):
        """Resets g to 0"""
        self.g = 0

    def optimise(self, weights, grad_weights):
        """
         Moves the weights in the direction opposite to weights gradient to attempt to reduce the Loss.

         AdaGrad is an extension of Stochastic gradient descent. AdaGrad allows for an adaptive learning rate by keeping
         a vector g that accumulates the square of grad_weights. As the grad_weights get smaller, so does the learning
         rate

         Parameters
         ----------
         weights : ndarray
             The existing weights which the optimiser needs to change.
         grad_weights : ndarray
             The weights gradient that determine which direction do the `weights` move

         Returns
         -------
         ndarray
             The new weights

         Notes
         -----
         g1 = g0 + (dL/dw)^2
         w1 = w0 - n(dL/dw)/(g1 + e)
         where:
             w1 = the new weights
             w0 = The existing weights
             g1 = the new g
             g0 = the existing g
             n = learning rate
             dL/dw = The weights gradient
             e = epsilon

         eplison is added to prevent division by 0

         """
        self.g = self.g + np.square(grad_weights)
        return weights - (self.learning_rate * grad_weights) / np.sqrt(self.g + 1E-07)



import numpy as np
from copy import deepcopy

from sinnn.Initialisers import SmallRandom, Zeros
from sinnn.Losses import CrossEntropy
from sinnn.Optimisers import AdaGrad
from sinnn.utils import batches, sigmoid
from sinnn import Metrics


class Model:
    """The neural network model"""

    def __init__(self, weights_initialiser=SmallRandom(), bias_initialiser=Zeros(),
                 loss_function=CrossEntropy(), optimiser=AdaGrad()):
        """
        Initialises Model parameters.

        Parameters
        ----------
        weights_initialiser : object
            the weights initialiser to be used. Instance of any class from module Initialsers may be used.
        bias_initialiser : object
            the bias initialiser to be used. Instance of any class from module Initialsers may be used.
        loss_function : object
            the loss function to be used. Instance of any class from module Losses may be used.
        optimiser: object
            the optimiser to be used. Instance of any class from module Optimisers may be used.

        """

        self.weights_initialiser = weights_initialiser
        self.bias_initialiser = bias_initialiser
        self.loss_function = loss_function
        self.optimiser = optimiser
        self.isbinary = None
        self.layers = []
        self.train_log = {}
        pass

    def add(self, *layers):
        """
        Adds a layer to the model

        Layers are stored in model.layers. Model parameters are passed to every class that doesn't already have those
        parameters defined at the level of the layer. This allows complete flexibility in setting layers with different
        parameters to each other.

        Parameters
        ----------
        layers : object
            the layer to be added. Instance of any class from module Layers may be used.

        """

        for layer in layers:
            if hasattr(layer, 'weights_initialiser'):
                if not layer.weights_initialiser:
                    layer.weights_initialiser = deepcopy(self.weights_initialiser)
            if hasattr(layer, 'bias_initialiser'):
                if not layer.bias_initialiser:
                    layer.bias_initialiser = deepcopy(self.bias_initialiser)
            if hasattr(layer, 'optimiser'):
                if not layer.optimiser:
                    layer.optimiser = deepcopy(self.optimiser)
            self.layers.append(layer)

    def reset(self):
        """Calls the reset method for all layers in model"""
        map(lambda layer: layer.reset(), self.layers)

    def logits(self, x):
        """
        Completes a forward pass through all the layers of the model to calculate logits.

        Parameters
        ----------
        x : ndarray
            2d input array. The first layer's input contain input of the model, while the latter layer's inputs are
            their previous layers outputs.

        Returns
        -------
        ndarray
            array of logits.

        """
        for index, layer in enumerate(self.layers):
            layer.forward(self.layers[index - 1].output) if index > 0 else layer.forward(x)
        return self.layers[-1].output

    def predict(self, x):
        """
        Completes a forward pass through all the layers of the model to calculate predictions.

        Parameters
        ----------
        x : ndarray
            2d input array. The first layer's input contain input of the model, while the latter layer's inputs are
            their previous layers outputs.

        Returns
        -------
        ndarray
            1d array of predicted labels.
        """
        return (np.round(sigmoid(self.logits(x)))
                if self.isbinary else
                np.argmax(self.logits(x), axis=1)) \
            if self.loss_function.__class__.__name__ == 'CrossEntropy' \
            else self.logits(x)

    def train(self, x_train, y_train, batch_size, epochs, validation=(), metrics=("loss",)):
        """
        Trains the model.

        Parameters
        ----------
        x_train: ndarray
            2d array containing train data of shape (number of examples, predictors per example)
        y_train: ndarray
            1d or 2d array containing train labels of shape (number of examples,) or (number of examples, 1)
        batch_size: int
            defines how many examples each batch should contain
        epochs : int
            defines how many epochs should the training include
        validation : tuple
            tuple of ndarrays of len 2 containing the validation inputs and labels.
        metrics : tuple
            tuple of strings defining what metrics should be computed by the model. Any function from module Metrics may
            be given.

        """

        #Exception handling
        if not len(x_train) == len(y_train):
            raise Exception("x_train and y_train must have the same length")
        if not len(x_train) % batch_size == 0:
            raise Exception("training set must be divisible by batch_size")

        # check if validation data has been provided.
        if len(validation) > 0:
            if not len(validation) == 2:
                raise Exception(f"Validation accepts exactly 2 arguments(x_val, y_val). {len(validation)} were given")
            if not len(validation[0]) == len(validation[1]):
                raise Exception("x_val and y_val must have the same length")
            else:
                x_val, y_val = validation[0], validation[1]
                val = True
        else:
            x_val, y_val = None, None
            val = False

        self.isbinary = (True if self.layers[-1].neurons == 1 else False)

        # If labels as 1d array, reshape to 2d array
        if self.isbinary:
            if self.loss_function.isbinary is None:
                self.loss_function.isbinary = True
            if y_train.ndim == 1:
                y_train = y_train.reshape(-1, 1)
            if val:
                if y_val.ndim == 1:
                    y_val = y_val.reshape(-1, 1)

        # Metrics reporting
        print(self.report_metrics(0, x_train, y_train, x_val, y_val, val, metrics))

        # training
        for i in range(epochs):

            # splitting into batches for training
            for x_batch, y_batch in batches(x_train, y_train, batch_size):
                loss_gradient = self.loss_function.loss_gradient(self.logits(x_batch), y_batch)

                for index, layer in enumerate(reversed(self.layers)):
                    loss_gradient = layer.backward(loss_gradient)

            # Metrics reporting
            print(self.report_metrics(i+1, x_train, y_train, x_val, y_val, val, metrics))

    def report_metrics(self, epoch, x_train, y_train, x_val, y_val, val, metrics):
        """
        Computes and reports metrics.

        Parameters
        ----------
        epoch : int
            which epoch is currently running
        x_train: ndarray
            2d array containing train data of shape (number of examples, predictors per example)
        y_train: ndarray
            1d or 2d array containing train labels of shape (number of examples,) or (number of examples, 1)
        x_val: ndarray
            2d array containing validation data of shape (number of examples, predictors per example)
        y_val: ndarray
            1d or 2d array containing validation labels of shape (number of examples,) or (number of examples, 1)
        val
        metrics

        Returns
        -------
        dict

        """
        m = dict()
        m["Epochs"] = epoch
        if epoch == 0 and self.train_log:
            self.train_log = {}

        logits = ([self.logits(x_train), self.logits(x_val)] if val else [self.logits(x_train)])
        labels = ([y_train, y_val] if val else [y_train])

        if type(metrics) == str:
            metrics = (metrics,)

        for metric in metrics:
            if metric not in self.train_log:
                self.train_log[metric] = {}
            keys = ([f"train_{metric}", f"validation_{metric}"] if val else [f"train_{metric}"])

            for i, key in enumerate(keys):
                if key not in self.train_log[metric]:
                    self.train_log[metric][key] = []
                met = getattr(Metrics, metric)(logits[i], labels[i], self.loss_function.__class__.__name__, self.isbinary)
                m[key] = met
                self.train_log[metric][key].append(met)

        return m



import numpy as np
import pickle


def sigmoid(x):
    """
    Applies the sigmoid transformation on a given 1d vector

    Parameters
    ----------
    x : ndarray
        1d Array to be sigmoided

    Returns
    -------
    ndarray
        1d Array of same shape as x containing sigmoided values of x

    Notes
    -----
    σ(x) = 1/(1 + e^-x) = (e^x)/(1 + e^x)
    where:
        σ = sigmoid function
        x = input
        e = Euler's number

    """
    return np.where(x >= 0, 1 / (1 + np.exp(-x)), np.exp(x) / (1 + np.exp(x)))


def softmax(x):
    """
    Applies the softmax transformation on a given 2d vector

    In order to make softmax more stable, x.max is subtracted from each value of x before performing softmax.

    Parameters
    ----------
    x : ndarray
        2d array to be softmaxed

    Returns
    -------
    ndarray
        2d Array of same shape as x containing softmaxed values of x

    Notes
    -----
    σ(x) = (e^x_i)/ Σ (e^x)
    where:
        σ = softmax function
        x_i = ith value of x
        x = input
        e = Euler's number

    """

    x = x - x.max(axis=1, keepdims=True)
    y = np.exp(x)
    return y / y.sum(axis=1, keepdims=True)


def one_hot(labels, categories):
    """
    One hot encodes the labels.

    Parameters
    ----------
    labels : ndarray
        1d array of ints containing labels.

    categories :  int
        the number of categories to encode the labels onto

    Returns
    -------
    ndarray
        2d Array of shape (len(x), categories) containing one hot encoded labels

    """
    y = np.zeros((labels.size, categories))
    y[np.arange(labels.size), labels] = 1
    return y


def batches(x, y, batch_size):
    """
    Yields successive batch_sized chunks of inputs and corresponding outputs from list.

    Parameters
    ----------
    x : list
        list of inputs
    y : list
        list of outputs
    batch_size :
        size of the batches to be yielded.

    Returns
    -------
    list
        batch of inputs of length (batch_size)
    list
        batch of outputs of length (batch_size)

    """

    for i in range(0, len(x), batch_size):
        yield x[i:i + batch_size], y[i:i + batch_size]


def save_model(model, filename='model'):
    """
    Saves an object to file

    Parameters
    ----------
    model : object
        any object to be stored on file
    filename : str
        name of the file the object needs to be stored on. Default: model

    """
    with open(filename, 'wb') as f:
        pickle.dump(model, f, protocol=-1)


def load_model(filename='model'):
    """
    Loads an object from file.

    Parameters
    ----------
    filename: str
        name of the file the object needs to be loaded from. Default: model

    Returns
    -------
        object that was stored on the file
    """
    with open(filename, 'rb') as f:
        return pickle.load(f)



import numpy as np


class ReLU:
    """Applies the Rectified Linear Unit transformation and backpropogates loss gradient"""

    def __init__(self):
        """Initialises ReLU."""

        self.input = None
        self.output = None
        pass

    def reset(self):
        """resets layer parameters"""
        self.input = None
        self.output = None

    def forward(self, x):
        """
        Applies the relu function on inputs and stores the results in outputs.

        Parameters
        ----------
        x : ndarray
            input of layer

        Notes
        -----
        output = (input if input > 0 else 0)

        """

        self.input = x
        self.output = np.maximum(self.input, 0.0)

    def backward(self, grad_output):
        """
        Backpropogates the loss gradient to calculate dL/dx.

        Relu layer does not have any weights so no need to calculate dL/dw.

        Parameters
        ----------
        grad_output : ndarray
            2d Array of same shape as `self.input` representing output gradient

        Returns
        -------
        ndarray
            2d Array of same shape as `self.input` representing input gradient

        Notes
        -----
        dL/dx = dL/dz * dz/dx
        dz/dx = (1 if input > 0 else 0)
            where:
            dL/dx = input gradient
            dL/dz = output gradient
            dz/dx = ReLU gradient

        """

        relu_grad = self.input > 0
        grad_input = grad_output * relu_grad
        return grad_input


class Dense:
    """Performs the transformation f(X) = X @ W and backpropogates loss gradient"""

    def __init__(self, neurons, weights_initialiser=None, bias_initialiser=None, optimiser=None):
        """
        Initialises layer parameters.

        Parameters
        ----------
        neurons : int
            number of neurons in layer
        weights_initialiser : object
            the weights initialiser to be used. Instance of any class from module Initialsers may be used.
        bias_initialiser : object
            the bias initialiser to be used. Instance of any class from module Initialsers may be used.
        optimiser: object
            the optimiser to be used. Instance of any class from module Optimisers may be used.
        """

        self.neurons = neurons
        self.weights_init = None
        self.biases_init = None
        self.weights = None
        self.input = None
        self.output = None
        self.weights_initialiser = weights_initialiser
        self.bias_initialiser = bias_initialiser
        self.optimiser = optimiser

    def init_weights(self, input_dim):
        """
        Initialise dense layer weights.

        Parameters
        ----------
        input_dim : int
            Number of columns in input matrix.

        """

        self.weights_init = self.weights_initialiser.initialise(input_dim, self.neurons)

    def init_biases(self):
        """Initialise dense layer biases"""

        self.biases_init = self.bias_initialiser.initialise(1, self.neurons)

    def reset(self):
        """Resets learnable layer parameters and initialises weights vector. Weights shape: [input_dim + 1, neurons]"""

        self.input = None
        self.output = None
        self.weights = np.vstack((self.biases_init, self.weights_init))
        self.optimiser.reset()

    def forward(self, x):
        """
        Calculates y = X @ W

        Parameters
        ----------
        x : ndarray
            input matrix of shape (batch_size, `input_dim` + 1)

        """

        # Initialising weights for first time use
        if self.weights is None:
            self.init_weights(x.shape[1])
            self.init_biases()
            self.reset()
        # adding a neuron/column of ones to support bias in weights matrix
        self.input = np.hstack((np.ones((x.shape[0], 1)), x))
        # self.output shape: (batch size, self.neurons)
        self.output = np.dot(self.input, self.weights)

    def backward(self, grad_output):
        """
        Backpropogates incoming gradient output to adjust layer weights and returns gradient input

        Parameters
        ----------
        grad_output : ndarray
            2d Array of shape (batch size, self.neurons)

        Returns
        -------
        ndarray
            2d Array of same shape as self.input representing input gradient

        Notes
        -----
        dL/dw = dz/dw * dL/dz
        dz/dw = x.T
        where:
            dL/dw = weights gradient
            dL/dz = output gradient
            x.T = inputs transposed
        dL/dx = dL/dz * dz/dx
        dz/dx = w.T
        where:
            dL/dx = input gradient
            dL/dz = output gradient
            w.T = weights transposed
        """
        # in calculatating grad_input, biases.T don't need to be multiplied to grad_output therefore are removed.
        grad_input = np.dot(grad_output, self.weights[1:, :].T)
        grad_weights = np.dot(self.input.T, grad_output)
        self.weights = self.optimiser.optimise(self.weights, grad_weights)
        return grad_input



import numpy as np
from sinnn.utils import sigmoid, softmax, one_hot


def loss(z, y, loss_function, isbinary=None):
    """
    Calculates the loss given the loss function used.

    Parameters
    ----------
    z : ndarray
        Array containing logits.
    y : ndArray
        Array containing labels.
    loss_function : str
        Name of loss function.
    isbinary : bool
        Whether the loss function should calculate Binary CrossEntropy or Categorical CrossEntropy

    Returns
    -------
    float64
        mean loss os of z from y

    Notes
    -----

    MSE:
        L = Σ 1/n (y - z)^2
        where:
            L = loss
            n = length of 1d array `z`
            y = labels
            z = logits

    CrossEntropy:

        Binary CrossEntropy:
        L = Σ 1/n -(ylog(σ(z)) + (1-y)log(1 - σ(z)))
        where:
            L = loss
            n = length of 1d array `z`
            y = labels
            log = natural logarithm
            σ = sigmoid function
            z = logits

        Categorical CrossEntropy:
        L = Σ 1/n Σ -(ylog(σ(z)))
        where:
            L = loss
            n = length of array `z`
            y = one hot encoded labels
            log = natural logarithm
            σ = softmax function
            z = logits

    """
    if loss_function == 'MSE':
        return np.mean(np.square(y - z))
    elif loss_function == 'CrossEntropy':
        return -np.mean(np.multiply(y, np.log(sigmoid(z))) + np.multiply(1 - y, np.log(1 - sigmoid(z)))) \
            if isbinary else \
            -np.mean(np.sum(np.multiply(one_hot(y, z.shape[1]), np.log(softmax(z))), axis=1))
    else:
        return None


def accuracy(z, y, loss_function, isbinary):
    """
    Calculates the accuracy if loss_function is CrossEntropy.

    For binary labels, `z` will be sigmoided to normalise their values and then rounded to compare to labels to
    calculate accuracy. For categorical labels, the index of the output neuron that gives the largest value is compared
    to labels to calculate accuracy. None is returned if loss_function does not support accuracy.

    Parameters
    ----------
    z : ndarray
        Array containing logits.
    y : ndArray
        Array containing labels.
    loss_function : str
        Name of loss function to verify if accuracy is supported.
    isbinary : bool
        Whether the accuracy should be calculated for Binary or Categorical labels.

    Returns
    -------
    float64
        accuracy of z with respect to y if `loss_function` supports accuracy else None.

    """
    return ((np.mean(np.round(sigmoid(z)) == y)
             if isbinary else
             np.mean(np.argmax(z, axis=1) == y))
            if loss_function == 'CrossEntropy' else
            None)
